from rltk.io.reader import BlockReader


class BlockArrayReader(BlockReader):
    """
    Block reader which input is an array.
    
    Args:
        array (list): A Python list which stores blocks.
    """

    def __init__(self, array):
        if not isinstance(array, list):
            raise TypeError('Type of array should be list.')
        self._array = array

    def __next__(self):
        for block in self._array: # iterate json objects
            for id1, id2s in block.items(): # iterate on one json object
                for id2 in id2s:
                    yield id1, id2



import dbm

from rltk.io.adapter import KeyValueAdapter
from rltk.io.serializer import Serializer, PickleSerializer
from rltk.record import Record


class DBMAdapter(KeyValueAdapter):
    """
    Python builtin `DBM <https://docs.python.org/3.6/library/dbm.html>`_ adapter.
    
    Args:
        filename (str): DBM file name.
        dbm_class (dbm): The value can be `dbm.gnu`, `dbm.ndbm` or `dbm.dumb`.
        serializer (Serializer, optional): The serializer used to serialize Record object. 
                                If it's None, `PickleSerializer` will be used. Defaults to None.
        
    Note:
        Performance drops when dataset is large.
    """
    def __init__(self, filename, dbm_class=dbm.ndbm, serializer: Serializer = None):
        if not serializer:
            serializer = PickleSerializer()
        self._db = dbm_class.open(filename, 'c')
        self._serializer = serializer

        # not all dbm supports iterator, simulate an iterator by getting a key set
        # key will be in byte form once read out from dbm, needs to be converted to string
        self._ids = set([k.decode('utf-8') for k in self._db.keys()])

    def get(self, record_id):
        return self._serializer.loads(self._db.get(record_id))

    def set(self, record_id, record: Record):
        self._ids.add(record_id)
        self._db[record_id] = self._serializer.dumps(record)

    def __next__(self):
        for i in self._ids:
            yield self.get(i)

    def close(self):
        self._db.close()



from rltk.record import Record


class KeyValueAdapter(object):
    """
    Super class of adapter of key value stores.
    """
    def __init__(self):
        pass

    def __del__(self):
        """
        Same as :meth:`close`.
        """
        self.close()

    def get(self, record_id) -> Record:
        """
        Get record.
        
        Args:
            record_id (str): Record id.
            
        Returns:
            Record:
        """
        raise NotImplementedError

    def set(self, record_id, record: Record):
        """
        Set record.
        
        Args:
            record_id (str): Record id.
            record (Record): Record object.
        """
        raise NotImplementedError

    def __iter__(self):
        """
        Same as :meth:`__next__`.
        """
        return self.__next__()

    def __next__(self):
        """
        Iterator of the data store. This is not required.
        
        Returns:
            iter: record_id, record
        """
        pass

    def close(self):
        """
        Close handler if needed.
        """
        pass


from rltk.blocking.block_generator import BlockGenerator
from rltk.blocking.custom_block_generator import CustomBlockGenerator
from rltk.blocking.inverted_index_block_generator import InvertedIndexBlockGenerator
from rltk.blocking.canopy_block_generator import CanopyBlockGenerator
from rltk.blocking.blocking_helper import BlockingHelper



import sys

from rltk.io.writer import BlockWriter


class BlockArrayWriter(BlockWriter):
    """
    Block writer stores data in array (list).
    
    Args:
        set_size (int, optional): Maximum size of a block. Defaults to `sys.maxsize`.
        index_blacklist (set, optional): If a Record's id is in blacklist, it won't be in output. Default's to None.
    """
    def __init__(self, set_size: int = sys.maxsize, index_blacklist: set = None):
        self._array = list()
        self._dict = dict()
        self._is_dirty = False
        self._set_size = set_size
        self._blacklist = index_blacklist or set()

    def write(self, id1: str, id2: str):
        if id1 in self._blacklist:
            return

        self._dict[id1] = self._dict.get(id1, set())
        s = self._dict[id1].add(id2)
        self._is_dirty = True

        if len(self._dict[id1]) > self._set_size:
            self._blacklist.add(id1)
            del self._dict[id1]

    def flush(self):
        if not self._is_dirty:
            return
        self._array = [{id1: list(id2s)} for id1, id2s in self._dict.items()]
        self._is_dirty = False

    def get_handler(self):
        self.flush()
        return self._array



from rltk.io.writer.block_writer import BlockWriter
from rltk.dataset import Dataset


class BlockGenerator(object):
    """
    Super class of block generator
    
    Args:
        dataset1 (Dataset): Dataset 1.
        dataset2 (Dataset): Dataset 2.
        writer (BlockWriter): Block writer.
        **kwargs: Key word arguments used by concrete class.
    """

    def __init__(self, dataset1: Dataset, dataset2: Dataset, writer: BlockWriter, **kwargs):
        self._writer = writer
        self._dataset1 = dataset1
        self._dataset2 = dataset2
        self._kwargs = kwargs

    def generate(self):
        """
        Generate blocks (:meth:`_generate_blocks`) and return handler.
        
        Returns:
            obj: Block handler, which can be used by corresponding BlockReader.
            
        """

        self._generate_blocks()

        handler = self._writer.get_handler()
        self._writer.close()
        return handler

    def _generate_blocks(self):
        """
        Generate blocks. It needs to be overwritten.
        """
        raise NotImplementedError



from rltk.io.writer import Writer


class BlockWriter(Writer):
    """
    Super class of block writer
    """

    def write(self, id1: str, id2: str):
        """
        Args:
            id1 (str): Record 1 id.
            id2 (str): Record 2 id.
        """
        raise NotImplementedError

    def get_handler(self):
        """
        The handler which can be used as input of corresponding :meth:`BlockReader`.
        """
        raise NotImplementedError

    def flush(self):
        """
        Force to flush buffer.
        """
        raise NotImplementedError

    def get_blacklist(self):
        """
        Black list of indices. 
        """
        if getattr(self, '_blacklist'):
            return self._blacklist
        raise NotImplementedError

    def close(self):
        """
        Close handler.
        """
        self.flush()
        super().close()



import json
import math
import numpy as np
import random
import itertools

from rltk.blocking import BlockGenerator


class CanopyBlockGenerator(BlockGenerator):
    """
    Canopy based block generator.

    Args:
        t1 (float): The loose distance.
        t2 (float): The tight distance.
        vectorize_function (Callable): Convert values in record to a vector. 
                                    The signature is `vectorize(r: Record) -> List`.
        distance_metric (Callable): Compute the distance between two vectors.
                                    The signature is `distance(v1: List, v2: List) -> float`
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._t1 = self._kwargs.get('t1')
        self._t2 = self._kwargs.get('t2')
        self._vectorize_function = self._kwargs.get('vectorize_function')
        if self._t1 <= self._t2:
            raise ValueError('t1 should be greater than t2')
        if self._t1 <= 0 or self._t2 <= 0:
            raise ValueError('t1 and t2 should greater than 0')
        self._distance_metric = self._kwargs.get('distance_metric')

    def _generate_blocks(self):
        vec_id_mapping = {}  # map data to
        dataset = []
        for r1 in self._dataset1:
            v = self._vectorize_function(r1)
            k = self._encode_key(v)
            vec_id_mapping[k] = vec_id_mapping.get(k, [])
            vec_id_mapping[k].append((0, r1.id))  # 0 means dataset1
            dataset.append(v)
        for r2 in self._dataset2:
            v = self._vectorize_function(r2)
            k = self._encode_key(v)
            vec_id_mapping[k] = vec_id_mapping.get(k, [])
            vec_id_mapping[k].append((1, r2.id))  # 1 means dataset2
            dataset.append(v)

        clusters = self._run_canopy_clustering(dataset, self._t1, self._t2, self._distance_metric)
        for c in clusters:
            id1s, id2s = [], []
            for vec in c:
                ids = vec_id_mapping[self._encode_key(vec)]
                for ds_id, r_id in ids:
                    if ds_id == 0:
                        id1s.append(r_id)
                    else:
                        id2s.append(r_id)
            for id1, id2 in itertools.product(id1s, id2s):
                self._writer.write(id1, id2)

    @staticmethod
    def _encode_key(v):
        return json.dumps(v)

    @staticmethod
    def _run_canopy_clustering(dataset, t1, t2, distance_metric):
        """
        The algorithm proceeds as follows, using two thresholds t1 (the loose distance) and t2 (the tight distance), 
        where t1 > t2.

        1. Begin with the set of data points to be clustered.
        2. Remove a point from the set, beginning a new 'canopy' containing this point.
        3. For each point left in the set, assign it to the new canopy \
            if its distance to the first point of the canopy is less than the loose distance t1.
        4. If the distance of the point is additionally less than the tight distance t2, 
            remove it from the original set.
        5. Repeat from step 2 until there are no more data points in the set to cluster.
        """
        canopies = []
        while len(dataset) > 0:
            center_idx = random.randint(0, len(dataset) - 1)
            center_vec = dataset[center_idx]
            new_canopy = []
            delete_list = []
            del dataset[center_idx]

            for d_idx in range(len(dataset)):
                d = dataset[d_idx]
                distance = distance_metric(center_vec, d)
                if distance < t1:
                    new_canopy.append(d)
                if distance < t2:
                    delete_list.append(d_idx)

            # delete vector from dataset from backward
            for d_idx in sorted(delete_list, reverse=True):
                del dataset[d_idx]
            new_canopy.append(center_vec)  # add center
            canopies.append(new_canopy)
        return canopies



import json
import os
import sys

from rltk.io.writer import BlockWriter


class BlockFileWriter(BlockWriter):
    """
    Block writer stores data in file.
    
    Args:
        filename (str): File name.
        buffer_size (int, optional): Buffer size. Defaults to 10,000.
        set_size (int, optional): Maximum size of a block. Defaults to `sys.maxsize`.
        index_blacklist (set, optional): If a Record's id is in blacklist, it won't be in output. Default's to None.
    """
    def __init__(self, filename: str, buffer_size: int = 10000,
                 set_size: int = sys.maxsize, index_blacklist: set = None):
        self._filename = filename
        self._temp_filename = filename + '.temp'
        self._buffer_size = buffer_size
        self._data_size_in_buffer = 0
        self._dict = dict() # buffer
        self._set_size = set_size
        self._blacklist = index_blacklist or set()

        # clean up output file
        open(self._filename, 'w').close()

    def write(self, id1: str, id2: str):
        # skip if id1 is in blacklist
        if id1 in self._blacklist:
            return

        # add pairs
        self._dict[id1] = self._dict.get(id1, set())
        if id2 not in self._dict[id1]:
            self._dict[id1].add(id2)
            self._data_size_in_buffer += 1

        # update id1 to blacklist when reaching threshold and remove id1 in memory
        if len(self._dict[id1]) > self._set_size:
            self._blacklist.add(id1)
            self._data_size_in_buffer -= len(self._dict[id1])
            del self._dict[id1]

        # flush when buffer is full
        if self._data_size_in_buffer >= self._buffer_size:
            self.flush()

    def get_handler(self):
        self.flush()
        return self._filename

    def flush(self):
        if self._data_size_in_buffer == 0:
            return

        fp = open(self._filename, 'r')
        temp_fp = open(self._temp_filename, 'w')

        # concatenate old and new id2s
        for line in fp:
            old_obj = json.loads(line)
            if len(old_obj) != 1:
                raise TypeError('Invalid block file')

            id1 = list(old_obj.keys())[0]
            id2s = old_obj[id1]

            if id1 in self._blacklist:
                continue

            if id1 in self._dict:
                id2s = self._dict[id1] | set(id2s)
                del self._dict[id1]

            if len(id2s) >= self._set_size:
                self._blacklist.add(id1)
                # remove from buffer
                if id1 in self._dict:
                    del self._dict[id1]
                # remove from output
                continue

            temp_fp.write(json.dumps({id1: list(id2s)}) + '\n')

        # write new id1 and id2s
        for id1, id2s in self._dict.items():
            temp_fp.write(json.dumps({id1: list(id2s)}) + '\n')

        fp.close()
        temp_fp.close()
        self._dict = dict()
        self._data_size_in_buffer = 0

        # replace file
        os.remove(self._filename)
        os.rename(self._temp_filename, self._filename)



import json
import heapq
import random
from typing import Callable
import pandas as pd
from collections import OrderedDict
from operator import itemgetter
import copy

from rltk.utils import get_record_pairs
from rltk.io.reader import GroundTruthReader
from rltk.io.writer import GroundTruthWriter

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from rltk.dataset import Dataset


class GroundTruth(object):
    """
    Ground Truth
    
    Args:
        filename (string, optional): existing ground truth file name
    """
    ID1 = 'id1'
    ID2 = 'id2'
    LABEL = 'label'

    def __init__(self, filename: str = None):
        self._ground_truth_data = {}
        self._gt_id1s = set([])
        self._gt_id2s = set([])

        if filename:
            self.load(filename)

    def add_positive(self, id1: str, id2: str):
        """
        Add a positive ground truth. It's a syntactic sugar of :meth:`add_ground_truth`.

        Args:
            id1 (str): first id
            id2 (str): second id
        """
        self.add_ground_truth(id1, id2, True)

    def add_negative(self, id1: str, id2: str):
        """
        Add a negative ground truth. It's a syntactic sugar of :meth:`add_ground_truth`.

        Args:
            id1 (str): first id
            id2 (str): second id
        """
        self.add_ground_truth(id1, id2, False)

    def add_ground_truth(self, id1: str, id2: str, value: bool):
        """
        Add a pair to ground truth

        Args:
            id1 (str): first id
            id2 (str): second id
            value (bool): ground truth value
        """
        key = self.encode_ids(id1, id2)
        self._ground_truth_data[key] = value
        self._gt_id1s.add(id1)
        self._gt_id2s.add(id2)

    def is_member(self, id1: str, id2: str) -> bool:
        """
        Check if this pair is in ground truth

        Args:
            id1 (str): first id
            id2 (str): second id

        Returns:
            bool:
        """
        key = self.encode_ids(id1, id2)
        return key in self._ground_truth_data

    def get_label(self, id1: str, id2: str) -> bool:
        """
        Pair's label
        
        Args:
            id1 (str): first id
            id2 (str): second id
        
        Returns:
            bool: True if positive, else negative
        """
        key = self.encode_ids(id1, id2)
        return self._ground_truth_data.get(key)

    def is_positive(self, id1: str, id2: str) -> bool:
        """
        If pair is positive in ground truth. It's a syntactic sugar of :meth:`get_label`.

        Args:
            id1 (str): first id
            id2 (str): second id

        Returns:
            bool:
            
        Raises:
            KeyError: if pair is not in ground truth
        """
        if not self.is_member(id1, id2):
            raise KeyError('Not in ground truth')
        return self.get_label(id1, id2)

    def is_negative(self, id1: str, id2: str) -> bool:
        """
        If pair is negative in ground truth. It's a syntactic sugar of :meth:`get_label`.

        Args:
            id1 (str): first id
            id2 (str): second id

        Returns:
            bool:
            
        Raises:
            KeyError: if pair is not in ground truth
        """
        if not self.is_member(id1, id2):
            raise KeyError('Not in ground truth')
        return not self.get_label(id1, id2)

    def load(self, filename: str):
        """
        Load the ground truth from file

        Args:
            filename (str): loading path
        """
        self.__init__()
        for obj in GroundTruthReader(filename):
            self._ground_truth_data[self.encode_ids(obj[self.ID1], obj[self.ID2])] = obj[self.LABEL] == 'True'

    def save(self, filename: str):
        """
        Save the ground truth to file

        Args:
            filename (str): saving path
        """
        w = GroundTruthWriter(filename)
        for k, v in self._ground_truth_data.items():
            ids = json.loads(k)
            w.write(ids[self.ID1], ids[self.ID2], v)
        w.close()

    def encode_ids(self, id1: str, id2: str):
        """
        Encode ids to key in ground truth dictionary

        Args:
            id1 (str): first id
            id2 (str): second id

        Returns:
            string:
        """
        key = json.dumps({self.ID1: id1, self.ID2: id2})
        return key

    def decode_ids(self, key: str):
        """
        Decode key in ground truth dictionary to ids

        Args:
            key (str):

        Returns:
            tuple:
                id1 (str)
                id2 (str)
        """
        obj = json.loads(key)
        return obj[self.ID1], obj[self.ID2]

    def __iter__(self):
        """
        Same as :meth:`__next__`
        """
        return self.__next__()

    def __next__(self):
        """
        Iterator
        
        Returns:
            iter: id1, id2, label
        """
        for k, v in self._ground_truth_data.items():
            id1, id2 = self.decode_ids(k)
            yield id1, id2, v

    def __len__(self):
        """
        Size of ground truth
        """
        return len(self._ground_truth_data)

    def generate_negatives(self, dataset1: 'Dataset', dataset2: 'Dataset',
                           score_function: Callable, num_of_negatives: int = -1, range_in_gt: bool = False):
        """
        Args:
            dataset1 (Dataset): Dataset 1.
            dataset2 (Dataset): Dataset 2.
            score_function (Callable): User function, inputs are two :meth:`rltk.record.Record` s, return is a float.
            num_of_negatives (int, optional): Number of negatives to generate. 
                                            Default is -1 which will generate same number of negatives to positives.
            range_in_gt (bool, optional): The negatives will be generated within the range of ids 
                                        in ground truth if it's True,
                                        otherwise range will be the cross product of two datasets. 
                                        Default is False.
        """
        num_of_negatives = len(self) if num_of_negatives == -1 else num_of_negatives
        max_heap = []

        for r1, r2 in get_record_pairs(dataset1, dataset2):
            if not self.is_member(r1.id, r2.id) and \
                    (not range_in_gt or (r1.id in self._gt_id1s and r2.id in self._gt_id2s)):
                s = score_function(r1, r2)
                heapq.heappush(max_heap, (s, r1.id, r2.id))
                if len(max_heap) > num_of_negatives:
                    heapq.heappop(max_heap)

        for d in max_heap:
            r1_id, r2_id = d[1], d[2]
            self.add_negative(r1_id, r2_id)

    def generate_all_negatives(self, dataset1: 'Dataset', dataset2: 'Dataset', range_in_gt: bool = False):
        """
        Args:
            dataset1 (Dataset): Dataset 1.
            dataset2 (Dataset): Dataset 2.
            range_in_gt (bool, optional): The negatives will be generated within the range of ids 
                                        in ground truth if it's True,
                                        otherwise range will be the cross product of two datasets. 
                                        Default is False.
        """
        for r1, r2 in get_record_pairs(dataset1, dataset2):
            if not self.is_member(r1.id, r2.id) and \
                    (not range_in_gt or (r1.id in self._gt_id1s and r2.id in self._gt_id2s)):
                self.add_negative(r1.id, r2.id)

    def generate_stratified_negatives(self, dataset1: 'Dataset', dataset2: 'Dataset',
                                      classify: Callable, num_of_strata: int, random_seed: int = None,
                                      num_of_negatives: int = -1, range_in_gt: bool = False):
        """
        Args:
            dataset1 (Dataset): Dataset 1.
            dataset2 (Dataset): Dataset 2.
            classify (Callable): User function, inputs are two :meth:`rltk.record.Record` s, 
                                return is an integer which identify which stratum the pair belongs to.
                                The return integer should be in range [0, num_of_strata).
            num_of_strata (int): Number of strata.
            random_seed (int, optional): The seed used by :py:meth:`random.seed`.
            num_of_negatives (int, optional): Number of negatives to generate. 
                                            Default is -1 which will generate same number of negatives to positives.
            range_in_gt (bool, optional): The negatives will be generated within the range of ids 
                                        in ground truth if it's True,
                                        otherwise range will be the cross product of two datasets. 
                                        Default is False.
        """

        # add positives and negatives to different clusters
        strata = [{'p': [], 'n': []} for _ in range(num_of_strata)]

        # build strata
        for r1, r2 in get_record_pairs(dataset1, dataset2):
            if range_in_gt and not (r1.id in self._gt_id1s and r2.id in self._gt_id2s):
                continue
            stratum_id = classify(r1, r2)
            p_n = 'p' if self.is_member(r1.id, r2.id) else 'n'
            strata[stratum_id][p_n].append((r1.id, r2.id))

        # compute weights: p / n
        strata_weights = {}
        for idx, s in enumerate(strata):
            stratum_id = str(idx)
            # nothing to pick
            if len(s['p']) == 0 or len(s['n']) == 0:
                strata_weights[stratum_id] = 0.0
                continue
            strata_weights[stratum_id] = float(len(s['p'])) / len(s['n'])

        # sorting
        sorted_strata_weights = OrderedDict(sorted(strata_weights.items(), key=itemgetter(1), reverse=True))

        # find out the number of negatives to pick from each stratum
        total_num = sum([len(s['p']) for s in strata]) if num_of_negatives == -1 else num_of_negatives
        num_to_pick_from_each_stratum = [0] * num_of_strata
        curr_strata_weights = copy.deepcopy(sorted_strata_weights)
        for stratum_id in sorted_strata_weights.keys():
            if total_num <= 0 or len(curr_strata_weights) == 0:
                break
            weight = sorted_strata_weights[stratum_id]
            idx = int(stratum_id)
            # normalize weights
            denominator = sum([w for w in curr_strata_weights.values()])
            num_to_pick_from_each_stratum[idx] = \
                min(round(total_num * weight / denominator), len(strata[idx]['n']))
            # prep for next round
            total_num -= num_to_pick_from_each_stratum[idx]
            curr_strata_weights.popitem(last=False)

        # pick negatives
        if random_seed:
            random.seed(random_seed)
        for idx, num in enumerate(num_to_pick_from_each_stratum):
            negs = random.sample(strata[idx]['n'], num)
            for n in negs:
                self.add_negative(n[0], n[1])

    def train_test_split(self, test_ratio: float = 0.2, random_seed: int = None):
        """
        Train test split
        
        Args:
            test_ratio (float, optional): The ratio of test from all ground truth data. Default is 0.2.
            random_seed (int, optional): The seed used by :py:meth:`random.seed`.
            
        Returns:
            tuple:
                train_gt (GroudTruth)
                test_gt (GroudTruth)
        """
        size = len(self)
        test_size = int(size * test_ratio)
        if random_seed:
            random.seed(random_seed)

        indices = [i for i in range(size)]
        random.shuffle(indices)
        test_indices = set(indices[:test_size])

        train_gt, test_gt = GroundTruth(), GroundTruth()
        for idx, (id1, id2, label) in enumerate(self):
            if idx in test_indices:
                test_gt.add_ground_truth(id1, id2, label)
            else:
                train_gt.add_ground_truth(id1, id2, label)

        return train_gt, test_gt

    def generate_dataframe(self, **kwargs):
        """
        Generate data frame
        
        Returns:
            pandas.DataFrame:
        """
        columns = ['id1', 'id2', 'label']
        table = []
        for id1, id2, label in self:
            table.append([id1, id2, label])
        return pd.DataFrame(table, columns=columns, **kwargs)



from rltk.io.reader.block_reader import BlockReader
from rltk.io.writer.block_writer import BlockWriter


class BlockingHelper(object):
    """
    It provides some useful blocking helper methods.
    
    Args:
        reader1 (BlockReader): BlockReader 1.
        reader2 (BlockReader): BlockReader 2.
    """

    def __init__(self, reader1: BlockReader, reader2: BlockReader):
        self._reader1 = reader1
        self._reader2 = reader2

    def union(self, writer: BlockWriter):
        """
        Union two blocks.
        
        Args:
            writer (BlockWriter): Block writer.
        """

        for id1, id2 in self._reader1:
            writer.write(id1, id2)
        for id1, id2 in self._reader2:
            writer.write(id1, id2)
        writer.close()



from rltk.io.adapter import KeyValueAdapter
from rltk.record import Record


class MemoryAdapter(KeyValueAdapter):
    """
    Basic in-memory (dict) adapter.
    """
    def __init__(self):
        self._records = dict()

    def get(self, record_id):
        return self._records.get(record_id)

    def set(self, record_id, record: Record):
        self._records[record_id] = record

    def __next__(self):
        for r in self._records.values():
            yield r



from rltk.io.reader import Reader
from rltk.io.adapter import KeyValueAdapter, MemoryAdapter
from rltk.record import Record, generate_record_property_cache, get_property_names

import pandas as pd
from typing import Callable


class Dataset(object):
    """
    A set of records.
    
    Args:
        reader (Reader, optional): Input reader.
        record_class (type(Record), optional): Sub class of Record.
        adapter (KeyValueAdapter, optional): Specify where to store indexed data. Defaults to MemoryAdapter.
        size (int, optional): Maximum size of records will be in Dataset, defaults to None, can be used when debugging.
        sampling_function (callable, optional): Sampling function, `raw_object` is the only parameter. \
                                                If it returns True, record instance will be created.
    Note:
        Set reader, record_class and adapter if new a Dataset needs to be generated.
        If Dataset is already generated and stored in a permanent adapter, only adapter needs to be provided.
        
    """
    def __init__(self, reader: Reader = None, record_class: type(Record) = None, adapter: KeyValueAdapter = None,
                 size: int = None, sampling_function: Callable = None):
        if not adapter:
            adapter = MemoryAdapter()
        self._adapter = adapter
        self._size = size

        # build index
        if reader and record_class:
            self._reader = reader
            self._record_class = record_class
            self._sampling_function = sampling_function

            self._build_index()

    def _build_index(self):
        if not self._reader or not self._record_class:
            raise ValueError('Reader or Record class is not provided.')
        size = 0
        for raw_object in self._reader:
            if not self._sampling_function or self._sampling_function(raw_object):
                record_instance = self._record_class(raw_object)
                generate_record_property_cache(record_instance)
                self._adapter.set(record_instance.id, record_instance)

                size += 1
                if self._size and size >= self._size:
                    break

    def get_record(self, record_id):
        """
        Getter of a record.
        
        Args:
            record_id (str): Record id.
            
        Returns:
            Record: Record object.
        """
        return self._adapter.get(record_id)

    def generate_dataframe(self, **kwargs):
        """
        Generate Pandas Dataframe
        
        Returns:
            pandas.Dataframe
        """
        table = []
        columns = None

        # construct table
        for r in self:

            # generate columns based on first record
            if not columns:
                columns = get_property_names(r.__class__)

            # get data
            row_data = []
            for prop_name in columns:
                row_data.append(getattr(r, prop_name))

            # append data
            table.append(row_data)

        return pd.DataFrame(table, columns=columns, **kwargs)

    def __iter__(self):
        """
        Same as :meth:`__next__`
        """
        return self.__next__()

    def __next__(self):
        """
        Iterator
        
        Returns:
            iter: Record
        """
        for r in self._adapter:
            yield r



from rltk.io.reader import Reader


class BlockReader(Reader):
    """
    Block reader.
    """

    def __next__(self):
        """
        Iterator of id pairs generated according to blocks.
        
        Returns:
            iter: id1, id2.
        """
        raise NotImplementedError
